@@ -26,6 +26,78 @@
 #include "numlock.h"
 
 
+#ifdef HAVE_SHADOW
+#include <shadow.h>
+#endif
+
+#ifdef USE_PAM
+#include <string>
+
+int conv(int num_msg, const struct pam_message **msg,
+         struct pam_response **resp, void *appdata_ptr){
+    *resp = (struct pam_response *) calloc(num_msg, sizeof(struct pam_response));
+    Panel* panel = *static_cast<Panel**>(appdata_ptr);
+    int result = PAM_SUCCESS;
+    for (int i=0; i<num_msg; i++){
+        resp[i]->resp=0;
+        resp[i]->resp_retcode=0;
+        switch(msg[i]->msg_style){
+            case PAM_PROMPT_ECHO_ON:
+                // We assume PAM is asking for the username
+                panel->EventHandler(Panel::Get_Name);
+                switch(panel->getAction()){
+                    case Panel::Suspend:
+                    case Panel::Halt:
+                    case Panel::Reboot:
+                        resp[i]->resp=strdup("root");
+                        break;
+
+                    case Panel::Console:
+                    case Panel::Exit:
+                    case Panel::Login:
+                        resp[i]->resp=strdup(panel->GetName().c_str());
+                        break;
+                }
+                break;
+
+            case PAM_PROMPT_ECHO_OFF:
+                // We assume PAM is asking for the password
+                switch(panel->getAction()){
+                    case Panel::Console:
+                    case Panel::Exit:
+                        // We should leave now!
+                        result=PAM_CONV_ERR;
+                        break;
+
+                    default:
+                        panel->EventHandler(Panel::Get_Passwd);
+                        resp[i]->resp=strdup(panel->GetPasswd().c_str());
+                        break;
+                }
+                break;
+
+            case PAM_ERROR_MSG:
+            case PAM_TEXT_INFO:
+                // We simply write these to the log
+                // TODO: Maybe we should simply ignore them
+                cerr << APPNAME << ": " << msg[i]->msg << endl;
+                break;
+        }
+        if (result!=PAM_SUCCESS) break;
+    }
+    if (result!=PAM_SUCCESS){
+        for (int i=0; i<num_msg; i++){
+            if (resp[i]->resp==0) continue;
+            free(resp[i]->resp);
+            resp[i]->resp=0;
+        };
+        free(*resp);
+        *resp=0;
+    };
+    return result;
+}
+#endif
+
 extern App* LoginApp;
 
 void CatchSignal(int sig) {
@@ -53,17 +125,22 @@ void User1Signal(int sig) {
 }
 
 
-App::App(int argc, char** argv) {
-
+#ifdef USE_PAM
+App::App(int argc, char** argv):
+    pam(conv, static_cast<void*>(&LoginPanel)){
+#else
+App::App(int argc, char** argv){
+#endif
     int tmp;
     ServerPID = -1;
     testing = false;
     mcookie = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa";
+    Dpy = NULL;
 
     // Parse command line
     while((tmp = getopt(argc, argv, "vhp:d?")) != EOF) {
         switch (tmp) {
-        case 'p':	// Test theme
+        case 'p':    // Test theme
             testtheme = optarg;
             testing = true;
             if (testtheme == NULL) {
@@ -71,14 +148,14 @@ App::App(int argc, char** argv) {
                 exit(ERR_EXIT);
             }
             break;
-        case 'd':	// Daemon mode
+        case 'd':    // Daemon mode
             daemonmode = true;
             break;
-        case 'v':	// Version
+        case 'v':    // Version
             std::cout << APPNAME << " version " << VERSION << endl;
             exit(OK_EXIT);
             break;
-        case '?':	// Illegal
+        case '?':    // Illegal
             cerr << endl;
         case 'h':   // Help
             cerr << "usage:  " << APPNAME << " [option ...]" << endl
@@ -100,7 +177,6 @@ App::App(int argc, char** argv) {
 
 
 void App::Run() {
-
     DisplayName = DISPLAY;
 
 #ifdef XNEST_DEBUG
@@ -111,6 +187,7 @@ void App::Run() {
     }
 #endif
 
+
     // Read configuration and theme
     cfg = new Cfg;
     cfg->readConf(CFGFILE);
@@ -133,6 +210,20 @@ void App::Run() {
         }
     }
 
+#ifdef USE_PAM
+    try{
+        pam.start("slim");
+        pam.set_item(PAM::Authenticator::TTY, DisplayName);
+        pam.set_item(PAM::Authenticator::Requestor, "root");
+        pam.set_item(PAM::Authenticator::Host, "localhost");
+
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+        exit(ERR_EXIT);
+    };
+#endif
+
     bool loaded = false;
     while (!loaded) {
         themedir =  themebase + themeName;
@@ -169,10 +260,10 @@ void App::Run() {
 
 #ifndef XNEST_DEBUG
         OpenLog();
-		
-		if (cfg->getOption("daemon") == "yes") {
-			daemonmode = true;
-		}
+        
+        if (cfg->getOption("daemon") == "yes") {
+            daemonmode = true;
+        }
 
         // Daemonize
         if (daemonmode) {
@@ -202,7 +293,6 @@ void App::Run() {
     Scr = DefaultScreen(Dpy);
     Root = RootWindow(Dpy, Scr);
 
-
     // for tests we use a standard window
     if (testing) {
         Window RealRoot = RootWindow(Dpy, Scr);
@@ -219,9 +309,8 @@ void App::Run() {
     LoginPanel = new Panel(Dpy, Scr, Root, cfg, themedir);
 
     // Start looping
-    XEvent event;
     int panelclosed = 1;
-    int Action;
+    Panel::ActionType Action;
     bool firstloop = true; // 1st time panel is shown (for automatic username)
 
     while(1) {
@@ -239,54 +328,125 @@ void App::Run() {
             LoginPanel->OpenPanel();
         }
 
-        Action = WAIT;
-        LoginPanel->GetInput()->Reset();
+        LoginPanel->Reset();
         if (firstloop && cfg->getOption("default_user") != "") {
-            LoginPanel->GetInput()->SetName(cfg->getOption("default_user") );
+            LoginPanel->SetName(cfg->getOption("default_user") );
             firstloop = false;
         }
 
-        while(Action == WAIT) {
-            XNextEvent(Dpy, &event);
-            Action = LoginPanel->EventHandler(&event);
-        }
 
-        if(Action == FAIL) {
+        if (!AuthenticateUser()){
             panelclosed = 0;
             LoginPanel->ClearPanel();
             XBell(Dpy, 100);
-        } else {
-            // for themes test we just quit
-            if (testing) {
-                Action = EXIT;
-            }
-            panelclosed = 1;
-            LoginPanel->ClosePanel();
+            continue;
+        }
+        
 
-            switch(Action) {
-                case LOGIN:
-                    Login();
-                    break;
-                case CONSOLE:
-                    Console();
-                    break;
-                case REBOOT:
-                    Reboot();
-                    break;
-                case HALT:
-                    Halt();
-                    break;
-                case SUSPEND:
-                    Suspend();
-                    break;
-                case EXIT:
-                    Exit();
-                    break;
-            }
+        Action = LoginPanel->getAction();
+        // for themes test we just quit
+        if (testing) {
+            Action = Panel::Exit;
+        }
+        panelclosed = 1;
+        LoginPanel->ClosePanel();
+
+        switch(Action) {
+            case Panel::Login:
+                Login();
+                break;
+            case Panel::Console:
+                Console();
+                break;
+            case Panel::Reboot:
+                Reboot();
+                break;
+            case Panel::Halt:
+                Halt();
+                break;
+            case Panel::Suspend:
+                Suspend();
+                break;
+            case Panel::Exit:
+                Exit();
+                break;
         }
     }
 }
 
+#ifdef USE_PAM
+bool App::AuthenticateUser(void){
+    // Reset the username
+    try{
+        pam.set_item(PAM::Authenticator::User, 0);
+        pam.authenticate();
+    }
+    catch(PAM::Auth_Exception& e){
+        switch(LoginPanel->getAction()){
+            case Panel::Exit:
+            case Panel::Console:
+                return true; // <--- This is simply fake!
+            default:
+                break;
+        };
+        cerr << APPNAME << ": " << e << endl;
+        return false;
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+        exit(ERR_EXIT);
+    };
+    return true;
+}
+#else
+bool App::AuthenticateUser(void){
+    LoginPanel->EventHandler(Panel::Get_Name);
+    switch(LoginPanel->getAction()){
+        case Panel::Exit:
+        case Panel::Console:
+            cerr << APPNAME << ": Got a special command (" << LoginPanel->GetName() << ")" << endl;
+            return true; // <--- This is simply fake!
+        default:
+            break;
+    };
+    LoginPanel->EventHandler(Panel::Get_Passwd);
+    
+    char *encrypted, *correct;
+    struct passwd *pw;
+
+    switch(LoginPanel->getAction()){
+        case Panel::Suspend:
+        case Panel::Halt:
+        case Panel::Reboot:
+            pw = getpwnam("root");
+            break;
+        case Panel::Console:
+        case Panel::Exit:
+        case Panel::Login:
+            pw = getpwnam(LoginPanel->GetName().c_str());
+            break;
+    }
+    endpwent();
+    if(pw == 0)
+        return false;
+
+#ifdef HAVE_SHADOW
+    struct spwd *sp = getspnam(pw->pw_name);    
+    endspent();
+    if(sp)
+        correct = sp->sp_pwdp;
+    else
+#endif        // HAVE_SHADOW
+        correct = pw->pw_passwd;
+
+    if(correct == 0 || correct[0] == '\0')
+        return true;
+
+    encrypted = crypt(LoginPanel->GetPasswd().c_str(), correct);
+    return ((strcmp(encrypted, correct) == 0) ? true : false);
+}
+#endif
+
 
 int App::GetServerPID() {
     return ServerPID;
@@ -294,34 +454,104 @@ int App::GetServerPID() {
 
 // Hide the cursor
 void App::HideCursor() {
-	if (cfg->getOption("hidecursor") == "true") {
-		XColor		    black;
-		char		    cursordata[1];
-		Pixmap		    cursorpixmap;
-		Cursor		    cursor;
-		cursordata[0]=0;
-		cursorpixmap=XCreateBitmapFromData(Dpy,Root,cursordata,1,1);
-		black.red=0;
-		black.green=0;
-		black.blue=0;
-		cursor=XCreatePixmapCursor(Dpy,cursorpixmap,cursorpixmap,&black,&black,0,0);
-		XDefineCursor(Dpy,Root,cursor);
-	}
+    if (cfg->getOption("hidecursor") == "true") {
+        XColor            black;
+        char            cursordata[1];
+        Pixmap            cursorpixmap;
+        Cursor            cursor;
+        cursordata[0]=0;
+        cursorpixmap=XCreateBitmapFromData(Dpy,Root,cursordata,1,1);
+        black.red=0;
+        black.green=0;
+        black.blue=0;
+        cursor=XCreatePixmapCursor(Dpy,cursorpixmap,cursorpixmap,&black,&black,0,0);
+        XDefineCursor(Dpy,Root,cursor);
+    }
 }
 
 void App::Login() {
     struct passwd *pw;
     pid_t pid;
 
-    pw = LoginPanel->GetInput()->GetPasswdStruct();
+#ifdef USE_PAM
+    try{
+        pam.open_session();
+        pw = getpwnam(static_cast<const char*>(pam.get_item(PAM::Authenticator::User)));
+    }
+    catch(PAM::Cred_Exception& e){
+        // Credentials couldn't be established
+        cerr << APPNAME << ": " << e << endl;
+        return;
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+        exit(ERR_EXIT);
+    };
+#else
+    pw = getpwnam(LoginPanel->GetName().c_str());
+#endif
+    endpwent();
     if(pw == 0)
         return;
+    if (pw->pw_shell[0] == '\0') {
+        setusershell();
+        strcpy(pw->pw_shell, getusershell());
+        endusershell();
+    }
+
+    // Setup the environment
+    char* term = getenv("TERM");
+    string maildir = _PATH_MAILDIR;
+    maildir.append("/");
+    maildir.append(pw->pw_name);
+    string xauthority = pw->pw_dir;
+    xauthority.append("/.Xauthority");
+    
+#ifdef USE_PAM
+    // Setup the PAM environment
+    try{
+        if(term) pam.setenv("TERM", term);
+        pam.setenv("HOME", pw->pw_dir);
+        pam.setenv("SHELL", pw->pw_shell);
+        pam.setenv("USER", pw->pw_name);
+        pam.setenv("LOGNAME", pw->pw_name);
+        pam.setenv("PATH", cfg->getOption("default_path").c_str());
+        pam.setenv("DISPLAY", DisplayName);
+        pam.setenv("MAIL", maildir.c_str());
+        pam.setenv("XAUTHORITY", xauthority.c_str());
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+        exit(ERR_EXIT);
+    }
+#endif
 
     // Create new process
     pid = fork();
     if(pid == 0) {
+#ifdef USE_PAM
+        // Get a copy of the environment and close the child's copy
+        // of the PAM-handle.
+        char** child_env = pam.getenvlist();
+        pam.end();
+#else
+        const int Num_Of_Variables = 10; // Number of env. variables + 1
+        char** child_env = static_cast<char**>(malloc(sizeof(char*)*Num_Of_Variables));
+        int n = 0;
+        if(term) child_env[n++]=StrConcat("TERM=", term);
+        child_env[n++]=StrConcat("HOME=", pw->pw_dir);
+        child_env[n++]=StrConcat("SHELL=", pw->pw_shell);
+        child_env[n++]=StrConcat("USER=", pw->pw_name);
+        child_env[n++]=StrConcat("LOGNAME=", pw->pw_name);
+        child_env[n++]=StrConcat("PATH=", cfg->getOption("default_path").c_str());
+        child_env[n++]=StrConcat("DISPLAY=", DisplayName);
+        child_env[n++]=StrConcat("MAIL=", maildir.c_str());
+        child_env[n++]=StrConcat("XAUTHORITY=", xauthority.c_str());
+        child_env[n++]=0;
+#endif
+
         // Login process starts here
-        SwitchUser Su(pw, cfg, DisplayName);
+        SwitchUser Su(pw, cfg, DisplayName, child_env);
         string session = LoginPanel->getSession();
         string loginCommand = cfg->getOption("login_cmd");
         replaceVariables(loginCommand, SESSION_VAR, session);
@@ -332,7 +562,7 @@ void App::Login() {
             system(sessStart.c_str());
         }
         Su.Login(loginCommand.c_str(), mcookie.c_str());
-        exit(OK_EXIT);
+        _exit(OK_EXIT);
     }
 
 #ifndef XNEST_DEBUG
@@ -347,6 +577,7 @@ void App::Login() {
     }
     if (WIFEXITED(status) && WEXITSTATUS(status)) {
         LoginPanel->Message("Failed to execute login command");
+        sleep(3);
     } else {
          string sessStop = cfg->getOption("sessionstop_cmd");
          if (sessStop != "") {
@@ -355,7 +586,16 @@ void App::Login() {
         }
     }
 
-    // Close all clients
+#ifdef USE_PAM
+    try{
+        pam.close_session();
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+    };
+#endif
+
+// Close all clients
     KillAllClients(False);
     KillAllClients(True);
 
@@ -382,6 +622,15 @@ void App::Reboot() {
     // Stop alarm clock
     alarm(0);
 
+#ifdef USE_PAM
+    try{
+        pam.end();
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+    };
+#endif
+
     // Write message
     LoginPanel->Message((char*)cfg->getOption("reboot_msg").c_str());
     sleep(3);
@@ -398,6 +647,15 @@ void App::Halt() {
     // Stop alarm clock
     alarm(0);
 
+#ifdef USE_PAM
+    try{
+        pam.end();
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+    };
+#endif
+
     // Write message
     LoginPanel->Message((char*)cfg->getOption("shutdown_msg").c_str());
     sleep(3);
@@ -433,16 +691,26 @@ void App::Console() {
 
 
 void App::Exit() {
+#ifdef USE_PAM
+    try{
+        pam.end();
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+    };
+#endif
+
     if (testing) {
         char* testmsg = "This is a test message :-)";
         LoginPanel->Message(testmsg);
         sleep(3);
+        delete LoginPanel;
+        XCloseDisplay(Dpy);
     } else {
+        delete LoginPanel;
         StopServer();
         RemoveLock();
     }
-    XCloseDisplay(Dpy);
-    delete LoginPanel;
     delete cfg;
     exit(OK_EXIT);
 }
@@ -453,9 +721,18 @@ int CatchErrors(Display *dpy, XErrorEvent *ev) {
 }
 
 void App::RestartServer() {
-        StopServer(); 
-        RemoveLock();
-        Run();
+#ifdef USE_PAM
+    try{
+        pam.end();
+    }
+    catch(PAM::Exception& e){
+        cerr << APPNAME << ": " << e << endl;
+    };
+#endif
+
+    StopServer(); 
+    RemoveLock();
+    Run();
 } 
 
 void App::KillAllClients(Bool top) {
@@ -491,9 +768,9 @@ void App::KillAllClients(Bool top) {
 
 
 int App::ServerTimeout(int timeout, char* text) {
-    int	i = 0;
+    int    i = 0;
     int pidfound = -1;
-    static char	*lasttext;
+    static char    *lasttext;
 
     for(;;) {
         pidfound = waitpid(ServerPID, NULL, WNOHANG);
@@ -520,8 +797,8 @@ int App::ServerTimeout(int timeout, char* text) {
 
 
 int App::WaitForServer() {
-    int	ncycles	 = 120;
-    int	cycles;
+    int    ncycles     = 120;
+    int    cycles;
 
     for(cycles = 0; cycles < ncycles; cycles++) {
         if((Dpy = XOpenDisplay(DisplayName))) {
@@ -560,13 +837,6 @@ int App::StartServer() {
         } else if (pos == 0) {
             server[argc++] = args+pos;
         }
-        if (server[argc-1][0] == 'v' && server[argc-1][1] == 't') {
-            bool ok = false;
-            Cfg::string2int(server[argc-1]+2, &ok);
-            if (ok) {
-                hasVtSet = true;
-            }
-        }
         ++pos;
 
         if (argc+1 >= MAX_XSERVER_ARGS) {
@@ -576,6 +846,17 @@ int App::StartServer() {
             break;
         }
     }
+
+    for (int i=0; i<argc; i++) {
+        if (server[i][0] == 'v' && server[i][1] == 't') {
+            bool ok = false;
+            Cfg::string2int(server[i]+2, &ok);
+            if (ok) {
+                hasVtSet = true;
+            }
+        }
+    }
+
     if (!hasVtSet && daemonmode) {
         server[argc++] = "vt07";
     }
@@ -622,9 +903,9 @@ int App::StartServer() {
         NumLock::setOn();
     } else if (numlock == "off") {
         NumLock::setOff();
-	}
+    }
     
-	delete args;
+    delete args;
 
     return ServerPID;
 }
@@ -650,7 +931,7 @@ void App::StopServer() {
 
     // Catch X error
     XSetIOErrorHandler(IgnoreXIO);
-    if(!setjmp(CloseEnv))
+    if(!setjmp(CloseEnv) && Dpy)
         XCloseDisplay(Dpy);
 
     // Send HUP to process group
@@ -724,15 +1005,15 @@ void App::setBackground(const string& themedir) {
         } else if (bgstyle == "tile") {
             image->Tile(XWidthOfScreen(ScreenOfDisplay(Dpy, Scr)), XHeightOfScreen(ScreenOfDisplay(Dpy, Scr)));
         } else if (bgstyle == "center") {
-    	    string hexvalue = cfg->getOption("background_color");
+            string hexvalue = cfg->getOption("background_color");
             hexvalue = hexvalue.substr(1,6);
-    	    image->Center(XWidthOfScreen(ScreenOfDisplay(Dpy, Scr)), XHeightOfScreen(ScreenOfDisplay(Dpy, Scr)),
-        			    hexvalue.c_str());
+            image->Center(XWidthOfScreen(ScreenOfDisplay(Dpy, Scr)), XHeightOfScreen(ScreenOfDisplay(Dpy, Scr)),
+                        hexvalue.c_str());
         } else { // plain color or error
-    	    string hexvalue = cfg->getOption("background_color");
+            string hexvalue = cfg->getOption("background_color");
             hexvalue = hexvalue.substr(1,6);
-    	    image->Center(XWidthOfScreen(ScreenOfDisplay(Dpy, Scr)), XHeightOfScreen(ScreenOfDisplay(Dpy, Scr)),
-        			    hexvalue.c_str());
+            image->Center(XWidthOfScreen(ScreenOfDisplay(Dpy, Scr)), XHeightOfScreen(ScreenOfDisplay(Dpy, Scr)),
+                        hexvalue.c_str());
         }
         Pixmap p = image->createPixmap(Dpy, Scr, Root);
         XSetWindowBackgroundPixmap(Dpy, Root, p);
@@ -839,8 +1120,8 @@ string App::findValidRandomTheme(const string& set)
 
 
 void App::replaceVariables(string& input,
-			   const string& var,
-			   const string& value)
+               const string& var,
+               const string& value)
 {
     string::size_type pos = 0;
     int len = var.size();
@@ -851,19 +1132,19 @@ void App::replaceVariables(string& input,
 
 
 void App::CreateServerAuth() {
-	/* create mit cookie */
-	int i, r;
-	int hexcount = 0;
+    /* create mit cookie */
+    int i, r;
+    int hexcount = 0;
         string authfile;
-	string cmd;
-	char *digits = "0123456789abcdef";
+    string cmd;
+    char *digits = "0123456789abcdef";
         srand( time(NULL) );
-	for ( i = 0; i < 31; i++ ) {
-		r = rand()%16;
+    for ( i = 0; i < 31; i++ ) {
+        r = rand()%16;
                 mcookie[i] = digits[r];
                 if (r>9)
                         hexcount++;
-	}
+    }
         /* MIT-COOKIE: even occurrences of digits and hex digits */
         if ((hexcount%2) == 0) {
                 r = rand()%10;
@@ -871,12 +1152,12 @@ void App::CreateServerAuth() {
                 r = rand()%5+10;
         }
         mcookie[31] = digits[r];
-	/* reinitialize auth file */
-	authfile = cfg->getOption("authfile");
-	remove(authfile.c_str());
-        putenv(StrConcat("XAUTHORITY=", authfile.c_str()));
-        cmd = cfg->getOption("xauth_path") + " -q -f " + authfile + " add :0 . " + mcookie;
-        system(cmd.c_str());
+    /* reinitialize auth file */
+    authfile = cfg->getOption("authfile");
+    remove(authfile.c_str());
+    putenv(StrConcat("XAUTHORITY=", authfile.c_str()));
+    cmd = cfg->getOption("xauth_path") + " -q -f " + authfile + " add :0 . " + mcookie;
+    system(cmd.c_str());
 }
 
 char* App::StrConcat(const char* str1, const char* str2) {
@@ -889,8 +1170,8 @@ char* App::StrConcat(const char* str1, const char* str2) {
 void App::UpdatePid() {
     std::ofstream lockfile(cfg->getOption("lockfile").c_str(), ios_base::out);
     if (!lockfile) {
-	    cerr << APPNAME << ": Could not update lock file: " << cfg->getOption("lockfile").c_str() << std::endl;
-	    exit(ERR_EXIT);
+        cerr << APPNAME << ": Could not update lock file: " << cfg->getOption("lockfile").c_str() << std::endl;
+        exit(ERR_EXIT);
     }
     lockfile << getpid() << std::endl;
     lockfile.close();
